using System.ComponentModel;
using Nibi.Dal;
using Nibi.Negocio;
using System;
using System.Drawing;

namespace Negocio {

    public enum TipoEstado {
        Ninguno = 0,
        Pendiente = 1,
        Autorizado,
        Excluido
    }

    [Tabla("PilotosCarreras")]
    public class EPilotoCarrera : EIdBase, IColor{

        #region Constructores
        public EPilotoCarrera() : base() { }

        public EPilotoCarrera(int vId) : base(vId) { }
        #endregion

        #region Sobrecargas de EBase

        protected override void InicializarObjetoNuevo() {
            Dorsal = 0;
            Estado = TipoEstado.Pendiente;
            Motivo = null;
            HoraVerificacion = null;
            Piloto = null;
            Vehiculo = null;
            Copiloto = null;
        }

        protected override void GuardarCore() {
            using (AmbitoTransa transa = this.CrearAmbitoTransa("Guardar Inscrito (PilotoCarrera)")) {
                base.GuardarCore();
                CrearClasificaciones();
                if (mCampeonatosInscrito != null)
                    mCampeonatosInscrito.Guardar();
                ActualizarClasificaciones(false);
                transa.Commit();
            }
        }

        public override void Deshacer() {
            mClasificaciones = null;
            mCampeonatosInscrito = null;
            base.Deshacer();
        }

        public override bool EsMasculino {
            get { return true; } 
        }

        public override string NombreDeLaEntidad {
            get { return "PilotoCarrera"; }
        }

        public override bool Modificable(string nombreProp) {
            if (!EstaCargado) return true;

            switch (nombreProp) {
                default:
                    return true;
            }
        }

        [Browsable(false)]
        public override string NPToString {
            get { return NPDorsal; } 
        }
        #endregion

        #region Propiedades

        public const string NPDorsal = "Dorsal";
        private int mDorsal;

        [Campo, Requerido]
        public int Dorsal {
            get {
                PrePropertyRead(NPDorsal);
                return mDorsal;
            }
            set {
                PrePropertyChanged(NPDorsal);
                mDorsal = value;
                if (mCarrera != null && mCarrera.Tipo == TipoCarrera.Rally && DorsalesDuplicados(mDorsal)) {
                    AddError(new PropertyException( NPDorsal, this, "No puede haber dorsales duplicados"));
                } else {
                    RemoveError(NPDorsal);
                }
                OnPropertyChanged(NPDorsal);
            }
        }

        public const string NPEstado = "Estado";
        private TipoEstado mEstado;

        [Campo]
        public TipoEstado Estado {
            get {
                PrePropertyRead(NPEstado);
                return mEstado;
            }
            set {
                PrePropertyChanged(NPEstado);
                mEstado = value;
                //Si el estado es excluido, se comprueba si la carrera ha comenzado y si es as se lanza un error.
                //if (mEstado == TipoEstado.Excluido && this.Carrera != null && this.Carrera.CarreraComenzada) {
                //    AddError(new PropertyException(NPEstado, this, "No se puede excluir a un participante si la carrera ha comenzado."));
                //} else {
                //    RemoveError(NPEstado);
                //}
                OnPropertyChanged(NPEstado);
            }
        }

        public const string NPMotivo = "Motivo";
        private string mMotivo;

        [Campo(Tamanyo = 100)]
        public string Motivo {
            get {
                PrePropertyRead(NPMotivo);
                return mMotivo;
            }
            set {
                PrePropertyChanged(NPMotivo);
                mMotivo = value;
                OnPropertyChanged(NPMotivo);
            }
        }

        public const string NPHoraVerificacion = "HoraVerificacion";
        private DateTime? mHoraVerificacion;

        [Campo]
        public DateTime? HoraVerificacion {
            get {
                PrePropertyRead(NPHoraVerificacion);
                return mHoraVerificacion;
            }
            set {
                PrePropertyChanged(NPHoraVerificacion);
                mHoraVerificacion = value;
                OnPropertyChanged(NPHoraVerificacion);
            }
        }

        public const string NPPiloto = "Piloto";
        private EPiloto mPiloto;

        [Campo("PilotoId"), Requerido]
        public EPiloto Piloto {
            get {
                PrePropertyRead(NPPiloto);
                return mPiloto;
            }
            set {
                PrePropertyChanged(NPPiloto);
                mPiloto = value;
                if (mCopiloto != null && mPiloto != null && mCopiloto.Id == mPiloto.Id) {//son la misma persona
                    AddError(new PropertyException(NPPiloto, this, "No puede ser piloto y copiloto a la vez."));
                } else {
                    RemoveError(NPPiloto);
                    if (mPiloto != null && mCarrera != null && mCarrera.EstaInscrito(mPiloto)) {
                        AddError(new PropertyException(NPPiloto, this, "El piloto ya est inscrito en esta carrera."));
                    } else if (mPiloto != null && mCarrera != null && EstaInscritoEnMismaFecha(mPiloto)) {
                        RemoveError(NPPiloto, typeof(PropertyException));
                        AddError(new PropertyException(NPPiloto, this, "No puede ser inscrito por que ya tiene una inscripcin en otra carrera en las mismas fechas."));
                    } else {
                        RemoveError(NPPiloto, typeof(PropertyException));
                    }
                }
                OnPropertyChanged(NPPiloto);
            }
        }

        public const string NPCarrera = "Carrera";
        private ECarrera mCarrera;

        [Campo("CarreraId"), Requerido]
        public ECarrera Carrera {
            get {
                PrePropertyRead(NPCarrera);
                //mCarrera = GetCarrera();
                return mCarrera;
            }
            set {
                PrePropertyChanged(NPCarrera);
                mCarrera = value;
                if (mCarrera != null && mCarrera.NecesitaCopiloto() && mCopiloto == null) {
                    AddError(new PropertyException(NPCopiloto, this, "Tiene que elegir un copiloto."));
                } else {
                    RemoveError(NPCopiloto);
                }
                OnPropertyChanged(NPCarrera);
            }
        }

        public const string NPVehiculo = "Vehiculo";
        private EVehiculo mVehiculo;

        [Campo("VehiculoId"), Requerido]
        public EVehiculo Vehiculo {
            get {
                PrePropertyRead(NPVehiculo);
                return mVehiculo;
            }
            set {
                PrePropertyChanged(NPVehiculo);
                mVehiculo = value;
                OnPropertyChanged(NPVehiculo);
            }
        }

        public const string NPCopiloto = "Copiloto";
        private ECopiloto mCopiloto;

        [Campo("CopilotoId")]
        public ECopiloto Copiloto {
            get {
                PrePropertyRead(NPCopiloto);
                return mCopiloto;
            }
            set {
                PrePropertyChanged(NPCopiloto);
                mCopiloto = value;
                if (mCopiloto != null && mPiloto != null && mCopiloto.Id == mPiloto.Id) {//son la misma persona
                    AddError(new PropertyException(NPCopiloto, this, "No puede ser piloto y copiloto a la vez."));
                } else {
                    RemoveError(NPCopiloto);
                    if (mCopiloto != null && mCarrera != null && mCarrera.EstaInscrito(mCopiloto)) {
                        AddError(new PropertyException(NPCopiloto, this, "El copiloto ya est inscrito en esta carrera."));
                    } else if (mCopiloto != null && this.Carrera != null && EstaInscritoEnMismaFecha(mCopiloto)) {
                        RemoveError(NPCopiloto);
                        AddError(new PropertyException(NPCopiloto, this, "No puede ser inscrito por que ya tiene una inscripcin en otra carrera en las mismas fechas."));
                    } else {
                        RemoveError(NPCopiloto, typeof(PropertyException));
                    }
                }
                OnPropertyChanged(NPCopiloto);
            }
        }

        #endregion

        #region Sanciones
        public const string NPSanciones = "Sanciones";
        private EColeccion<ESancion> mSanciones = null;

        /// <summary>Devuelve una coleccin con todos las Sanciones.</summary>
        public EColeccion<ESancion> Sanciones {
            get {
                if(mSanciones == null) {
                    mSanciones = new EColeccion<ESancion>();
                    if(!this.EsNuevo) {
                        mSanciones.Filtros.Add(ESancion.NPPilotoCarrera, this);
                        mSanciones.ColumnaMostrarRango(new string[] {ESancion.NPTipoSancion, ESancion.NPMotivo, ESancion.NPImporte, ESancion.NPCronometrable, ESancion.NPTiempo});
                        mSanciones.Cargar();
                    }
                    mSanciones.AddingNew += new AddingNewEventHandler(mSanciones_AddingNew);
                }
                return mSanciones;
            }
        }

        private void mSanciones_AddingNew(object sender, AddingNewEventArgs e) {
            ESancion entidad = new ESancion();
            entidad.PilotoCarrera = this;
            e.NewObject = entidad;
        }

        #endregion

        #region Clasificaciones
        public const string NPClasificaciones = "Clasificaciones";
        private EColeccion<EClasificacionCarrera> mClasificaciones = null;

        /// <summary>Devuelve una coleccin con todos las Clasificaciones. Si es una clasificacin
        /// de un rally la primera entidad de la coleccin es la clasificacin total.</summary>
        public EColeccion<EClasificacionCarrera> Clasificaciones {
            get {
                if(mClasificaciones == null) {
                    mClasificaciones = new EColeccion<EClasificacionCarrera>();
                    if (Carrera.Tipo == TipoCarrera.Rally) {
                        ERally rally = new ERally(Carrera.Id);
                        if (rally.Existe) {
                            EClasificacionCarrera obj = GetClasificacion(rally);
                            if (obj != null) {
                                mClasificaciones.Add(obj);
                            }
                        }
                    }
                    foreach (ECronometrable cronometrable in Carrera.Cronometrables) {
                        if (cronometrable.Tipo != TipoCronometrable.Sector) {
                            EClasificacionCarrera obj = GetClasificacion(cronometrable);
                            if (obj != null) {
                                mClasificaciones.Add(obj);
                            }
                        }
                    }
                }

                return mClasificaciones;
            }
        }

        #endregion

        #region CampeonatosInscrito
        public const string NPCampeonatosInscrito = "CampeonatosInscrito";
        private EColeccion<ECampeonatoInscrito> mCampeonatosInscrito = null;

        /// <summary>Devuelve una coleccin con todos los CampeonatosInscrito.</summary>
        public EColeccion<ECampeonatoInscrito> CampeonatosInscrito {
            get {
                if(mCampeonatosInscrito == null) {
                    mCampeonatosInscrito = new EColeccion<ECampeonatoInscrito>();
                    if(!this.EsNuevo) {
                        mCampeonatosInscrito.Filtros.Add(ECampeonatoInscrito.NPPilotoCarrera, this);
                        mCampeonatosInscrito.Cargar();
                    }
                    mCampeonatosInscrito.AddingNew += new AddingNewEventHandler(mCampeonatosInscrito_AddingNew);
                }
                return mCampeonatosInscrito;
            }
        }

        private void mCampeonatosInscrito_AddingNew(object sender, AddingNewEventArgs e) {
            ECampeonatoInscrito entidad = new ECampeonatoInscrito();
            entidad.PilotoCarrera = this;
            e.NewObject = entidad;
        }

        #endregion

        #region TiemposPilotoSector
        public const string NPTiemposPilotoSector = "TiemposPilotoSector";
        private EColeccion<ETiempoPilotoSector> mTiemposPilotoSector = null;

        /// <summary>Devuelve una coleccin con todos los TiemposPilotoSector.</summary>
        public EColeccion<ETiempoPilotoSector> TiemposPilotoSector {
            get {
                if (!(Carrera != null && mCarrera.Tipo == TipoCarrera.Rally)) {
                    return null;
                }
                if(mTiemposPilotoSector == null) {
                    mTiemposPilotoSector = new EColeccion<ETiempoPilotoSector>();
                    if(!this.EsNuevo) {
                        mTiemposPilotoSector.Filtros.Add(ETiempoPilotoSector.NPPilotoCarrera, this);
                        mTiemposPilotoSector.Cargar();
                    }
                    mTiemposPilotoSector.AddingNew += new AddingNewEventHandler(mTiemposPilotoSector_AddingNew);
                    mTiemposPilotoSector.ListChanged += new ListChangedEventHandler(mTiemposPilotoSector_ListChanged);
                }
                return mTiemposPilotoSector;
            }
        }

        private void mTiemposPilotoSector_AddingNew(object sender, AddingNewEventArgs e) {
            ETiempoPilotoSector entidad = new ETiempoPilotoSector();
            entidad.PilotoCarrera = this;
            e.NewObject = entidad;
        }

        private void mTiemposPilotoSector_ListChanged(object sender, ListChangedEventArgs e) {
            if(!TiemposPilotoSector.EsValido)
                AddError(new PropertyException(NPTiemposPilotoSector, this, TiemposPilotoSector.AlgunError.Message, TiemposPilotoSector.AlgunError));
            else
                RemoveError(NPTiemposPilotoSector);
            OnPropertyCollectionChanged(NPTiemposPilotoSector);
        }
        #endregion

        #region IColor

        public Color DevolverColor() {
            return DevolverColor(mEstado);
        }

        public Color DevolverColor(object o) {
            TipoEstado estado;
            try {
                estado = (TipoEstado)o;
            } catch (Exception ex) {
                throw new InvalidCastException(ex.Message);
            }
            switch (estado) {
                case TipoEstado.Ninguno:
                    return Color.White;
                case TipoEstado.Pendiente:
                    return Color.CornflowerBlue;
                case TipoEstado.Autorizado:
                    return Color.LimeGreen;
                case TipoEstado.Excluido:
                    return Color.Red;
                //case TipoEstado.Descalificado:
                //    return Color.Tomato;
                default:
                    return Color.White;
            }
        }

        #endregion

        /// <summary>
        /// Devuelve true si hay otra carrera donde la instancia est inscrita y la carrera se encuentre entre las fechas de esta carrera.
        /// </summary>
        /// <param name="esb"></param>
        /// <returns></returns>
        private bool EstaInscritoEnMismaFecha(ESinClaveBase esb) {
            EPiloto piloto = esb as EPiloto;
            ECopiloto copiloto = esb as ECopiloto;
            EColeccion<ECarrera> colCarreras = ECarrera.GetCollection(true, false);
            foreach (ECarrera carrera in colCarreras) {
                if (!carrera.Equals(mCarrera) && (FechaEntre(this.mCarrera.FechaComienzo, carrera.FechaComienzo, carrera.FechaFin) || FechaEntre(this.mCarrera.FechaFin, carrera.FechaComienzo, carrera.FechaFin)) && ((piloto != null && carrera.EstaInscrito(piloto)) || (copiloto != null && carrera.EstaInscrito(copiloto)))) {
                    return true;
                }
            }
            return false;
        }

        private bool FechaEntre(DateTime valor, DateTime fechaIni, DateTime fechaFin) {
            //utilizamos el compare to
            if (valor.CompareTo(fechaIni) >= 0 && valor.CompareTo(fechaFin) <= 0) {
                return true;
            }
            return false;
        }

        private void CrearClasificaciones() {
            EClasificacionTotalRally clasifRally = null;
            ERally rally = new ERally(Carrera.Id);
            if (rally.Existe && rally.Tipo == TipoCarrera.Rally) {
                if (!TieneClasificacionRally(rally.ClasificacionesRally)) {//tiene que tener una clasificacionTotalRally, sino, se crea.
                    clasifRally = rally.ClasificacionesRally.AddNew();
                    clasifRally.PilotoCarrera = this;
                    clasifRally.Guardar();
                }
            }
            EColeccion<ECronometrable> colCronom = Carrera.Cronometrables;
            foreach (ECronometrable cronometrable in colCronom) {
                if (cronometrable.Tipo == TipoCronometrable.Manga) {
                    if (!TieneClasificacionCronometrable(cronometrable.ClasificacionesManga)) {
                        EClasificacionManga clasif = cronometrable.ClasificacionesManga.AddNew();
                        clasif.PilotoCarrera = this;
                        clasif.Guardar();
                    }
                } else if (cronometrable.Tipo == TipoCronometrable.Tramo && clasifRally != null) {
                    if (!TieneClasificacionCronometrable(cronometrable.ClasificacionesTramo)) {
                        EClasificacionTramo clasif = cronometrable.ClasificacionesTramo.AddNew();
                        clasif.PilotoCarrera = this;
                        clasif.ClasificacionRally = clasifRally;
                        clasif.Guardar();
                    }
                } else if (cronometrable.Tipo == TipoCronometrable.Sector) {
                    if (!TieneTiempoSector(cronometrable.TiemposPilotoSector)) {
                        ETiempoPilotoSector tiempoSector = cronometrable.TiemposPilotoSector.AddNew();
                        tiempoSector.PilotoCarrera = this;
                        tiempoSector.Guardar();
                    }
                }
            }
        }

        private void ActualizarClasificaciones(bool cargar) {
            EClasificacionTotalRally clasifRally = null;
            ERally rally = new ERally(Carrera.Id);
            if (rally.Existe && rally.Tipo == TipoCarrera.Rally) {
                if (cargar) {
                    rally.ClasificacionesRally.Cargar();
                }
                rally.ActualizarClasificacion();
                rally.ClasificacionesRally.Guardar();
            }
            EColeccion<ECronometrable> colCronom = Carrera.Cronometrables;
            foreach (ECronometrable cronometrable in colCronom) {
                if (cronometrable.Tipo == TipoCronometrable.Manga) {
                    if (cargar) {
                        cronometrable.ClasificacionesManga.Cargar();
                    }
                    cronometrable.ActualizarClasificacion();
                    cronometrable.ClasificacionesManga.Guardar();
                } else if (cronometrable.Tipo == TipoCronometrable.Tramo && clasifRally != null) {
                    if (cargar) {
                        cronometrable.ClasificacionesTramo.Cargar();
                    }
                    cronometrable.ActualizarClasificacion();
                    cronometrable.ClasificacionesTramo.Guardar();
                }
            }
        }

        private bool TieneTiempoSector(EColeccion<ETiempoPilotoSector> eColeccion) {
            if (eColeccion != null) {
                foreach (ETiempoPilotoSector tiempoSector in eColeccion) {
                    if (tiempoSector.PilotoCarrera.Equals(this)) {
                        return true;
                    }
                }
            }
            return false;
        }

        private bool TieneClasificacionCronometrable(EColeccion<EClasificacionManga> eColeccion) {
            foreach (EClasificacionManga clasifManga in eColeccion) {
                if (clasifManga.PilotoCarrera.Equals(this)) {
                    return true;
                }
            }
            return false;
        }

        private bool TieneClasificacionCronometrable(EColeccion<EClasificacionTramo> eColeccion) {
            foreach (EClasificacionTramo clasifTramo in eColeccion) {
                if (clasifTramo.PilotoCarrera.Equals(this)) {
                    return true;
                }
            }
            return false;
        }

        private bool TieneClasificacionRally(EColeccion<EClasificacionTotalRally> eColeccion) {
            if (eColeccion != null) {
                foreach (EClasificacionCarrera clasificacion in eColeccion) {
                    if (clasificacion.PilotoCarrera.Equals(this)) {
                        return true;
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// Devuelve la clasificacin total del inscrito dado un rally
        /// </summary>
        /// <param name="rally"></param>
        /// <returns></returns>
        private EClasificacionCarrera GetClasificacion(ERally rally) {
            foreach (EClasificacionTotalRally obj in rally.ClasificacionesRally) {
                if (obj.PilotoCarrera.Equals(this)) {
                    return obj;
                }
            }
            return null;
        }

        /// <summary>
        /// Devuelve la clasificacion del inscrito de un cronometrable (Manga o Tramo).
        /// </summary>
        /// <param name="cronometrable"></param>
        /// <returns></returns>
        private EClasificacionCarrera GetClasificacion(ECronometrable cronometrable) {
            if (cronometrable.Tipo == TipoCronometrable.Manga) {
                foreach (EClasificacionManga obj in cronometrable.ClasificacionesManga) {
                    if (obj.PilotoCarrera.Equals(this)) {
                        return obj;
                    }
                }
            } else if (cronometrable.Tipo == TipoCronometrable.Tramo) {
                foreach (EClasificacionTramo obj in cronometrable.ClasificacionesTramo) {
                    if (obj.PilotoCarrera.Equals(this)) {
                        return obj;
                    }
                }
            }
            return null;
        }

        public override void Eliminar() {
            using (AmbitoTransa transa = this.CrearAmbitoTransa("Eliminar Inscrito (PilotoCarrera)")) {
                //hay q eliminar tambin los campeonatos y las clasificaciones.
                EliminarSanciones();
                EliminarCampeonatosInscrito();
                EliminarClasificaciones();
                EliminarTiemposPilotoSector();
                base.Eliminar();
                transa.Commit();
            }
            //Actualizar clasificaciones.
            ActualizarClasificaciones(true);
        }

        private void EliminarTiemposPilotoSector() {
            EColeccion<ETiempoPilotoSector> col = TiemposPilotoSector;
            if (col != null) {
                int indice = 0;
                while (indice < col.Count) {
                    ETiempoPilotoSector obj = col[indice];
                    obj.Eliminar();
                    indice = indice + 1;
                }
            }
        }

        private void EliminarSanciones() {
            int indice = 0;
            EColeccion<ESancion> col = Sanciones;
            while (indice < col.Count) {
                ESancion obj = col[indice];
                obj.Eliminar();
                indice = indice + 1;
            }
        }

        private void EliminarClasificaciones() {
            int indice = 0;
            EClasificacionTotalRally ultimaClasificacionAEliminar = null;
            EColeccion<EClasificacionCarrera> col = Clasificaciones;
            while (col.Count > indice) {
                EClasificacionCarrera obj = Clasificaciones[indice];
                switch (obj.TipoClasificacion) {
                    case TipoClasificacionCarrera.Manga:
                        EClasificacionManga objClasifManga = new EClasificacionManga(obj.Id);
                        objClasifManga.Actualizar();
                        if (objClasifManga != null) {
                            objClasifManga.Eliminar();
                        }
                        break;
                    case TipoClasificacionCarrera.Rally:
                        //EClasificacionTotalRally objClasifRally = new EClasificacionTotalRally(obj.Id);
                        //objClasifRally.Actualizar();
                        //if (objClasifRally != null) {
                        //    objClasifRally.Eliminar();
                        //}
                        ultimaClasificacionAEliminar = new EClasificacionTotalRally(obj.Id);
                        break;
                    case TipoClasificacionCarrera.Tramo:
                        EClasificacionTramo objClasifTramo = new EClasificacionTramo(obj.Id);
                        //objClasifTramo.ClasificacionRally = null;
                        objClasifTramo.Actualizar();
                        if (objClasifTramo != null) {
                            objClasifTramo.Eliminar();
                        }
                        break;
                    default:
                        throw new InvalidCastException("Error al eliminar, los tipos de clasificaciones no coinciden.");
                }
                indice = indice + 1;
            }
            if (ultimaClasificacionAEliminar != null) {
                ultimaClasificacionAEliminar.Eliminar();
            }
        }

        private void EliminarCampeonatosInscrito() {
            int indice = 0;
            EColeccion<ECampeonatoInscrito> col = CampeonatosInscrito;
            while (indice < col.Count) {
                ECampeonatoInscrito obj = col[indice];
                obj.Eliminar();
                indice = indice + 1;
            }
        }

        private bool DorsalesDuplicados(int dorsal) {
            EColeccion<EPilotoCarrera> col = this.Carrera.PilotosCarrera;
            int contador = 0;
            int indice = 0;
            while (contador < 2 && indice < col.Count) {
                if (col[indice].Dorsal == dorsal) {
                    contador = contador + 1;
                }
                indice = indice + 1;
            }
            return contador >= 2;
        }

        public override bool EstaSucio {
            get {
                return base.EstaSucio || (mCampeonatosInscrito != null && mCampeonatosInscrito.EstaSucio);
            }
        }
    }
}

